//this isn't actually correct, since it does jacobi iterations
// instead of gauss-seidel, would need to make sure array entries aren't
// over-written. also, the sign of their averaging operator is wrong.

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
// #include <sys/time.h>
#include <math.h>
#ifndef M_PI
#define M_PI 3.14159265
#endif

#define TILE_SIZE 4
/*
 * --- implement your kernel here ---
 */

void heatKernelCUDA(float* A, int n) {
};

// define macro OUTPUT to print input & output matrix
#define OUTPUT

void checkCUDAError(const char *msg);

void initMatrix(float *A, int n);
void dstMatrix(float *A, int n);
float* CUDA_initMatrix(float *A, int n);
void CUDA_freeMatrix(float *Ad);
void printMatrix(char* name, float *A, int n);

int main(int argc, char *argv[])
{
   float *A; /* arrays for heat equation */
   int n; /* n = matrix size */
   int m; /* m = iterations */
   
   clock_t clock1, clock2;
   float cpu_time = 0;

   if (argc < 3) {
      printf("Error: please specify matrix size and number of iterations: \n");
      printf("Usage: %s 10 5 \n", argv[0]);
      exit(1);      
   };
   
   /* read block size and number of blocks */
   n = atoi(argv[1]);
   m = atoi(argv[2]);

   if (n % TILE_SIZE != 0) {
      printf("Error: array size has to be a multiple of tile size %d \n",
              TILE_SIZE);
      exit(1);      
   };

   printf("Heat equation on grid of size %d (%d repeats): \n", n,m);

   /*
    * allocate and initialise matrices in host memory:
    */
   A = (float *) malloc((n+2)*(n+2)*sizeof(float));
   dstMatrix(A,n);
   
#ifdef OUTPUT
   printMatrix("A",A,n+2); 
#endif

   printf("Start heat equation:\n");
   clock1 = clock();
   cpu_time = 0.0;

   /* init Matrices */
   float* Ad = CUDA_initMatrix(A,n+2);

   /* perform heat kernel (m repeats) */
   for(int i=0;i<m;i++) {
      /*
       * --- add kernel invocation here
       */
       
#ifdef OUTPUT
   printf("Transfer array from GPU to CPU\n"); 
#endif
      // transfer array from GPU into CPU memory:
      int size = (n+2)*(n+2)*sizeof(float);
      cudaMemcpy(A,Ad, size, cudaMemcpyDeviceToHost);
         checkCUDAError("memory of array not transferred back");

#ifdef OUTPUT
   printMatrix("Result: ",A,n+2);
#endif
   };

   clock2 = clock();
   cpu_time += (clock2-clock1)/(float)CLOCKS_PER_SEC;

#ifdef OUTPUT
   printMatrix("Final Result: ",A,n+2);
#endif

   printf("Execution time: %f (CPU sec) \n",cpu_time);
   printf("Performance: %E Flop/s \n",m*pow(n,3.0)/cpu_time);

   cudaFree(Ad);

   free(A);
   
   return(0);
}

/* set Matrix values to zero */
void initMatrix(float *A, int n)
{
   int i,k;

   for (i=0; i<n; i++)
     for (k=0; k<n; k++)
	A[i*n+k] = 0;
}

/* initialise Matrix: discrete Sine Transform */
void dstMatrix(float *A, int n)
{
   int i,k;

   for (i=0; i<n; i++)
     for (k=0; k<n; k++)
	A[i*n+k] = sin( (i*k*M_PI)/n );
}

/* 
 * heat equation (idealized time step) 
 *  -> standard C implementation 
 */
void heateq(float *A, int n)
{
   int i,j;
   int N = n+2;

   for (i=1; i<=n; i++)
     for (j=1; j<=n; j++)
     {
        A[i*N+j] = A[i*N+j] - 0.25*(
            A[(i-1)*N+j] + A[(i+1)*N+j] +
            A[i*N+(j-1)] + A[i*N+(j+1)]
        );
     }
}



/* 
 * matrix initialization 
 */
__host__ 
float* CUDA_initMatrix(float *M, int n)
{
   int size = n*n*sizeof(float);
   float* Md;
   
   cudaMalloc((void**)&Md, size); 
      checkCUDAError("allocate memory for Matrix");

   cudaMemcpy(Md,M, size, cudaMemcpyHostToDevice);
      checkCUDAError("memory of A not transferred");

   return Md;
}

/* print Matrix */
void printMatrix(char* name, float *A, int n)
{
   int i,k;

   printf("Matrix %s (size %d)\n",name,n);
   for (i=0; i<n; i++) {
     for (k=0; k<n; k++)
	printf("%f ", A[i*n+k]);
     printf("\n");
   };
}

/*
 * helper function to check for errors in CUDA calls
 * source: NVIDIA
 */
void checkCUDAError(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err)
    {
        fprintf(stderr, "\nCuda error (%s): %s.\n", msg, cudaGetErrorString( err) );
        exit(-1);
    }
}
